/*
*******************************************************************************************
**					NAVAL AIR WARFARE CENTER AIRCRAFT DIVISION (NAWCAD)
**					Test and Training Programs Division 5.1.3.3
**					23029 Cedar Point Rd, Building 2611
**					Patuxent River, MD	20670
**
**	Name:			BC635PCIdrvr.c
**	Title:			Datum bc635pic device driver
**
**	Description:	LynxOS Device Driver for 
**					DATUM BC635 PCI Time and Frequency Processor
**
**	Public entry points for:
**		interrupt	bc635pci_intr()
**		nstall		bc635pci_install()
**		unistall	bc635pci_uninstall()
**		open		bc635pci_open()
**		close		bc635pci_close()
**		ioctl		bc635pci_ioctl()
**		select		bc635pci_select()
**		read		bc635pci_read()
**		write		bc635pci_write()
**
**		kernal thread
**					bc635pci_kt_intr_server()
**
**	Local (private) functions:
**		rdTmTv()		Read time from the board in binary seconds format
**		rdTime()		Read time from the board in days through seconds format
**		sendCommand()	Send command to the board to take action
**		decode_drm_ret() decodes the status returned from DRM calls
**
**  History:
**
**	-- DATE --	-- Name --	--------- Description -------------------------------------
**	22 OCT 01	TJ Lanzi	MS Visual Studio workspace/project created for LynxOS device
**							driver - based on the Datum supplied Solaris driver stfp.c
**
*******************************************************************************************
*/

// INCLUDE FILES
// #define	__LYNXOS			// to get kernel functions and constants

#include <conf.h>
#include <kernel.h>
#include <sys/file.h>
#include <errno.h>
#include <sys/drm_errno.h>
#include <sys/ioctl.h>	
#include <dldd.h>				// for dynamically loadable drivers
#include <pci_resource.h>
#include <bc635pci_io.h>

#include "BC635PCIinfo.h"

#define	DEVICE_FAIL (char *)SYSERR

enum {	BC_PCI_UNDEFINE,
		BC_PCI_INSTALL,
		BC_PCI_OPEN
};
/*-------------------------  DATA  STRUCTURES  ------------------------------*/

typedef struct 	{						// BC635 device registers
	unsigned long				timereq;			// time request (for time[] regs)
	unsigned long				eventreq;			// time request (for event[] regs)
	unsigned long				unlock;				// time capture lockout release
	unsigned long				reserved;
	unsigned long				control;			// misc. control register
	unsigned long				ack;				// acknowledge register
	unsigned long				mask;				// interrupt mask register
	unsigned long				intstat;			// interrupt status register
	unsigned long				minstrb;			// minor strobe time
	unsigned long				mijstrb;			// major strobe time
	unsigned long				reserved2[2];
	unsigned long				time[2];			// minor/major time hold 
	unsigned long				event[2];			// minor/major event
}	bc635pci_regs;


typedef struct {						// Driver Statistics Structure
	struct drm_node_s			*bc635pci_dev_h;	// drm device handle
	struct bc635pci_devinfo		*dip;				// device info
	// DRM initialization variables
	int							bus_number;			
	int							device_number;
	int							func_number;
	// variables for the Kernel thread function
	int							thread_id;
	int							thread_prio;
	int							thread_intr_queue;
	int							thread_run;

	unsigned int				cmdregval;
	unsigned long				*vaddr0;
	unsigned long				*vaddr1;

	bc635pci_regs				*pRegs;				// Device control regs
	unsigned char				*pDpr;				// DPRAM base pointer
	unsigned char 				*pIn;				// usr input area in DPRAM
	unsigned char				*pOut;				// output area (to user) in DPRAM
	unsigned char				*pGps;				// GPS packet area in DPRAM
	unsigned char				*pYear;				// year area in DPRAM
	unsigned int				dpramSize;			// size of DPRAM
	unsigned long				intServiced;		// last interrupt(s) seviced
	unsigned long				intMaskSig;			// Interrupt mask for signals
	unsigned long				state;
	int				 			mutex;				// mutex for protecting critical code
	int							cv;					// condition variable
	int							proc_id;			// process ID for signals
	int							key;				// used for chaining interrupts
}	bc635pci_statics;

// debug variables - memory that can be checked by skdb
	unsigned long	BC635_isr_enter = 0;
	unsigned long	BC635_unclaimed = 0;
	unsigned long	BC635_k_thread = 0;


/*------------------------ INTERRUPT HANDLER --------------------------------*/

// ------------------------------------------------------------------------ 
// Name: bc635pci_intr
// 
// Description: 
// 
// low level interrupt handler
// 
// Input: pointer
// 
// Returns: 
// ------------------------------------------------------------------------ 


void	bc635pci_intr(bc635pci_statics *psu)
{
int	cnt;

	cnt = BC635_isr_enter;
	BC635_isr_enter = cnt + 1;

// TJL 6-MAY-02		move code to kernel thread so that the mutex can be used
//					this also make most efficient use of interrupt time.
	ssignal (&psu->thread_intr_queue);

	ioint_link(psu->key);
}

// ------------------------------------------------------------------------ 
// Name: bc635pci_kt_intr_server
// 
// Description: 
// 
// 		handles the service of the interrupt via a kernel thread
//		this solves the problem of the swait (mutex_enter) that should
//		not be in the ISR
// 
// Input: pointer
// 
// Returns: 
// ------------------------------------------------------------------------ 

void	bc635pci_kt_intr_server (bc635pci_statics *psu)
{

unsigned long		pending;

// bc635pci_intr dispatches this thread

// thread initialization

	psu->thread_run = 1;
	// debug variable 
	BC635_k_thread = 1;	// thread started

//	loop forever

//	swait - on the semaphore
	while (psu->thread_run)
	{
		swait (&psu->thread_intr_queue, SEM_SIGRETRY);
#ifdef	DEBUG_DRVR
	cprintf ("bc635pci_kt_intr_server ()\tINTR = %d \tUNCLM = %d \tTHRD = %d\n"
										,BC635_isr_enter ,BC635_unclaimed ,BC635_k_thread);

#endif
		// debug variable 
		BC635_k_thread +=1;	// thread started

	// ISSUE - blocking on the mutex  if this thread is higher prio then user tasks
	// another possible soluction is to disable and restore the interupts in place of
	// the mutex
		swait(&psu->mutex, SEM_SIGIGNORE);					// mutex_enter()

		pending = (psu->pRegs->mask & psu->pRegs->intstat & INT_BITMASK);

		/* claim or reject interrupt check */

		if(pending != 0)
		{
			psu->intServiced |= pending & psu->intMaskSig;
			psu->pRegs->intstat = pending;

			if(pending & INT_PACKET)
						psu->pRegs->ack = ACK_PACKET;
			// the LynxOS method
			if ((psu->proc_id != 0) && 
				(pending & psu->intMaskSig)) 
			{
				if (_kill(psu->proc_id, SIGUSR1) != OK) 
				{
					// assume process ID was invalid
					psu->proc_id = 0;
				}
			}

			if(pending & ~psu->intMaskSig & INT_PACKET)
				ssignal(&psu->cv);					// signal condition

		}
		else
		{
			BC635_unclaimed += 1;
		}

		ssignal(&psu->mutex);					// mutex_exit()
	}	// end while (psu->thread_run);			// loop forever

}




/*------------------------ USER FUNCTIONS -----------------------------------*/


/*
****************************************************************************
* rdTmTv - read time from the board in binary seconds format
****************************************************************************
*/
static	void	rdTimeTv (	bc635pci_statics		*psu,
							unsigned long			*pTmReg, 
							struct stfp_timeval		*pTmT)
{
	pTmT->tv.tv_sec		= *(pTmReg+1);
	pTmT->tv.tv_usec	= *(pTmReg) & 0xfffff;
	pTmT->hnsec			= (*(pTmReg) >> 20) & 0xf;
	pTmT->status		= (*(pTmReg) >> 24) & 0xf;
}

/*
****************************************************************************
* rdTime - read time from the board in days through seconds format
****************************************************************************
*/
static	void	rdTime (	bc635pci_statics		*psu,
							unsigned long			*pTmReg,
							unsigned char			*pY,
							struct stfp_time		*pTmS)
{
	pTmS->tm.tm_year =	(*(pY) << 8) + (*(pY+1) - TM_YEAR_BASE);
	pTmS->tm.tm_yday =	((*(pTmReg) >> 20)&0xf00) + ((*(pTmReg+1) >> 24)&0xff) - 1;
	pTmS->tm.tm_hour =	(*(pTmReg+1) >> 16) & 0xff;
	pTmS->tm.tm_min =	(*(pTmReg+1) >> 8) & 0xff;
	pTmS->tm.tm_sec =	*(pTmReg+1) & 0xff;
	pTmS->usec =		*(pTmReg) & 0xfffff;
	pTmS->hnsec =		(*(pTmReg) >> 20) & 0xf;
	pTmS->status =		(*(pTmReg) >> 24) & 0xf;
}

/*
****************************************************************************
* sendCommand - send command to the board to take action
****************************************************************************
*/
static	int sendCommand	( bc635pci_statics		*psu)
{
	int i;
	unsigned long	delay;
#define	DELAY_COUNT		20100
#define	USEC_DELAY_CNT	50

#ifdef DEBUG_DRVR
	cprintf("sendCommand() ENTER\n");
#endif

	psu->pRegs->ack = ACK_COMMAND;			// clear ack command done bit
	usec_sleep (5);							// TJL put a delay between the two writes
	psu->pRegs->ack = ACK_ACTION;			// force bc635PCI to take action 

#if 0	// old wait loop
	for (i = 0;
		((psu->pRegs->ack & ACK_COMMAND) == 0) && (i < 20);
				i++)
	{
		delay = (1 << i);
		usec_sleep (delay);
	}
#else	// new wait loop
#endif
	for (i = 0;
		((psu->pRegs->ack & ACK_COMMAND) == 0) && (i < DELAY_COUNT);
				i++)
	{
		usec_sleep (USEC_DELAY_CNT);
	}

	if(i >= DELAY_COUNT)							/* check if timeout occured */
	{
#ifdef DEBUG_DRVR
 cprintf ("sendCommand() timeout occured waiting to send data packet\n");
#endif
		pseterr (EIO);
		return(SYSERR);
	}

#ifdef DEBUG_DRVR
	cprintf("sendCommand() EXIT\n");
#endif
	
	return (0);
}

// ------------------------------------------------------------------------ 
// Name: decode_drm_ret
// 
// Description: 
// 
// decode the return status - 
// 
// Input:	int value representing value return from drm call
//			char string for debug purposes
// 
// Returns: pointer to the statics structure
// ------------------------------------------------------------------------ 

static	void	decode_drm_ret (int	drm_ret,	char *text)
{


		switch (drm_ret)
		{
		case DRM_ENODEV:
			cprintf("bc635pci ERROR! No Device\n");
			break;
		case	DRM_EFAULT:
			cprintf("bc635pci ERROR! DRM_EFAULT Bad Handle Pointer\n");
			break;
		case	DRM_EINVALID:
			cprintf("bc635pci ERROR! DRM_EINVALID\n");
			break;
		case	DRM_EMEMMAP:
			cprintf("bc635pci ERROR! DRM_EMEMMAP\n");
			break;
		case	DRM_EBADSTATE:
			cprintf("bc635pci ERROR! DRM_EBADSTATE\n");
			break;
		default:
			cprintf("bc635pci ERROR! %d Mapping BAR1\n", drm_ret);
			break;
		}
}

    
// ------------------------------------------------------------------------ 
// Name: bc635pci_install
// 
// Description: 
// 
// The install entry point is the entry point which is invoked by LynxOS when 
// the device driver is first installed in the system. (through the devinstall 
// utility)
// 
// Input: devinfo structure pointer
// 
// Returns: pointer to the statics structure
// ------------------------------------------------------------------------ 

char * bc635pci_install(struct bc635pci_devinfo *info)
{
// TJL 30-APR-02 The device info should be a binary file NOT the object module
	bc635pci_statics		*psu;
	int						drm_ret;

#define	PCI_MEM_ENABLE 0x02


#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() Entry Point\n");
#endif

// allocate (memory for) the device info structure
	psu = (bc635pci_statics *)sysbrk(sizeof(bc635pci_statics));
	psu->dip = info;	// save device info ptr
	psu->pDpr = NULL;
	psu->pRegs = NULL;
	psu->bc635pci_dev_h = NULL;
	psu->state = BC_PCI_UNDEFINE;
	psu->thread_intr_queue = 0;		// semaphore for kernel thread - bc635pci_kt_intr_server
	psu->mutex = 1;					// initialize mutex
	pi_init (&psu->mutex);			// TJL 30 APR 02 correct way to init semaphore
	psu->cv	= 0;					// initialize condition to zero - no events occurred
	psu->bus_number = 0;
	psu->device_number = 0;
	psu->func_number = 0;
	psu->cmdregval = 0;

	// Initialize the debug variables to zero
	BC635_isr_enter = 0;
	BC635_unclaimed = 0;
	BC635_k_thread = 0;

#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() bc635pci_statics allocated\n");
#endif

	//	Claims access to the specified device
	drm_ret = drm_get_handle(	PCI_BUSLAYER,				// Bus layer (PCI_BUSLAYER)
						BC_PCI_DEFAULT_VENDOR_ID,	// BC635PCI Device Vendor ID
						BC_PCI_DEFAULT_DEVICE_ID,	// BC635PCI PCI Device ID
						&psu->bc635pci_dev_h);		// Handle to the device node
	if (drm_ret != DRM_OK) 
	{
		decode_drm_ret (drm_ret, "drm_get_handle()");
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}


#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() drm_get_handle to PCI device complete\n");
#endif


	/* The bc635PCI is divided into three register fields:
	** the Boot EPROM field, the dual-port RAM field and the device register field
	** The Boot EPROM field begins at address zero within the card's physical
	** address space.  The starting address and length of the two remaining spaces should
	** be obtained from the 'reg' attribute 
	/*
	** map the base address register 0 -
	** The offset of each register is relative to the start of the 
	** device register field starting address found in BAR0
	*/
	// based on the sample solaris code it looks like the DPRAM is mapped to BAR1
	drm_ret = drm_map_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1, &psu->vaddr1);
	if ( drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "Mapping BAR1");
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}

	drm_ret = drm_map_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0, &psu->vaddr0);
	if ( drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "Mapping BAR0");
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}

	psu->pDpr = (unsigned char *)psu->vaddr1;	// from BAR1
	psu->pRegs = (bc635pci_regs *)psu->vaddr0;	// BAR0

	// from chapter 12 (Sample program) Device Resource Manager
	// enable the device for memory access...
	drm_ret = drm_device_read (psu->bc635pci_dev_h, PCI_RESID_CMDREG, 0, 0,
			(void *)&psu->cmdregval);
	if ( drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "drm_device_read(PCI_RESID_CMDREG)");
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0);
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}

	psu->cmdregval |= PCI_MEM_ENABLE;
	drm_ret = drm_device_write (psu->bc635pci_dev_h, PCI_RESID_CMDREG, 0, 0,
			(void *)&psu->cmdregval);
	if ( drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "drm_device_write(PCI_RESID_CMDREG)");
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0);
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}
	// these reads are here for informational purposes only as far as I can tell
	
	drm_ret = drm_device_read (psu->bc635pci_dev_h, PCI_RESID_BUSNO, 0, 0,
		(void *)&psu->bus_number);
	if (drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "drm_device_read(PCI_RESID_BUSNO)");
	}
	drm_ret = drm_device_read (psu->bc635pci_dev_h, PCI_RESID_DEVNO, 0, 0,
		(void *)&psu->device_number);
	if (drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "drm_device_read(PCI_RESID_DEVNO)");
	}
	
	drm_ret = drm_device_read (psu->bc635pci_dev_h, PCI_RESID_FUNCNO, 0, 0,
		(void *)&psu->func_number);
	if (drm_ret != DRM_OK)
	{
		decode_drm_ret (drm_ret, "drm_device_read(PCI_RESID_FUNCNO)");
	}

	
#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() pRegs = %x \n", psu->pRegs);
	cprintf("bc635pci_install() pDpr  = %x \n", psu->pDpr);
#endif	// end debug

/*	Mapping the Dual Ported Ram (DPRAM) Areas
**	
**	The DPRAM is byte wide , the current size is 2048 bytes
**	There are four areas in DPRAM to map to
**
**	Input Area		used for sending commands to the TFP
**	Output Area		holds data requested by user
**	GPS Packet Area	holds packets of data from GPS reciever
**	Year Area		holds the year number
**	
**	The offset of each DPRAM area(relative to the DPRAM starting address)
**	is stored at the top of the DPRAM where TOP is the last DPRAM address.
**----------------------------------------------------------------------
**	TOP_DPRAM = 2048 
**	Input Area Offset LSB			Base Address + TOP_DPRAM - 1
**	Input Area Offset MSB			Base Address + TOP_DPRAM - 2
**	Output Area Offset LSB			Base Address + TOP_DPRAM - 3
**	Output Area Offset MSB			Base Address + TOP_DPRAM - 4
**	GPS Packet Area Offset LSB		Base Address + TOP_DPRAM - 5
**	GPS Packet Area Offset MSB		Base Address + TOP_DPRAM - 6
**	Year Area Offset LSB			Base Address + TOP_DPRAM - 7
**	Year Area Offset MSB			Base Address + TOP_DPRAM - 8
**
**	the virtual address of the DPRAM was mapped and stored in psu->pDpr
*/
	if (!recoset())
	{
		// Input Area
		psu->pIn	= psu->pDpr + ((psu->pDpr[TOP_DPRAM - 2]<< 8) + psu->pDpr[TOP_DPRAM - 1]);
		// Output Area
		psu->pOut	= psu->pDpr + ((psu->pDpr[TOP_DPRAM - 4]<< 8) + psu->pDpr[TOP_DPRAM - 3]); 
		// GPS Packet Area
		psu->pGps	= psu->pDpr + ((psu->pDpr[TOP_DPRAM - 6]<< 8) + psu->pDpr[TOP_DPRAM - 5]); 
		// Year Area
		psu->pYear	= psu->pDpr + ((psu->pDpr[TOP_DPRAM - 8]<< 8) + psu->pDpr[TOP_DPRAM - 7]);
	}
	else
	{
		cprintf ("Error Accessing DPRAM\n"); 
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0);
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}
	noreco();

	if ((psu->pIn - psu->pDpr) == 0x102) 
	{
		cprintf("\nSan Jose Div bc635/637PCI TimeCode Rec/Gen.  Revsion 3.1\n");
		cprintf("Copyright (c) 1998-2000 Datum, Inc.  all rights reserved\n");
	}
	else
	{
		cprintf("\nbc635pci ERROR! Did Not Boot Correctly - Try system warm reboot\n");
#ifdef DEBUG_DRVR
		cprintf("bc635pci_install() pIn = %x - pDpr = %x = %x\n", psu->pIn, psu->pDpr,
																(psu->pIn - psu->pDpr));
#endif

		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0);
		drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);
		drm_free_handle (psu->bc635pci_dev_h);
		sysfree(psu, sizeof(bc635pci_statics));
		pseterr (ENXIO);	// device does not exist.
		return (DEVICE_FAIL);
	}

	// register (install) low level interrupt handler
#ifdef	DEBUG_DRVR
	cprintf("bc635pci_install() drm_register_isr low level interrupt handler \n");
#endif
	
	psu->key = drm_register_isr (psu->bc635pci_dev_h, bc635pci_intr, psu);

	// create the kernel thread here
	psu->thread_prio = (_getpriority() * 2) + 1;

#ifdef	DEBUG_DRVR
	cprintf ("bc635pci_install() prio = %d starting the kernel thread\n", psu->thread_prio);
#endif
	psu->thread_id = ststart (	bc635pci_kt_intr_server,	// kernel thread
					256,				// stacksize
					psu->thread_prio,	// priority
					"bc635pci_kt_intr",	// name
					1,					// number of arguments
					psu);				// arg = device statistics
#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() %d = bc635pci_kt_intr_server() kernel thread started\n",
			psu->thread_id);
#endif


	psu->state = BC_PCI_INSTALL;
#ifdef DEBUG_DRVR
	cprintf("bc635pci_install() Exit Point - OK\n");
#endif
	return ((char *)psu);
}


// ------------------------------------------------------------------------ 
// Name: bc635pci_uninstall
// 
// Description: 
// The uninstall routine is called when the driver is uninstalled.
// 
// Input: statics structure pointer
// 
// Returns: nothing
// ------------------------------------------------------------------------
int	bc635pci_uninstall(bc635pci_statics *psu)
{

	if (psu == NULL)
	{
		pseterr (ENXIO);	// device does not exist.
		return(SYSERR);
	}
	// remove the kernel thread
	stremove (psu->thread_id);
	
	//	remove interrupt
	drm_unregister_isr (psu->bc635pci_dev_h);
	
	// remove interrupts here 
	psu->pRegs->mask = 0;
	
	// unmap resources
	drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR0);
	drm_unmap_resource (psu->bc635pci_dev_h, PCI_RESID_BAR1);

	//	free the device
	drm_free_handle (psu->bc635pci_dev_h);

	//  free memory
	sysfree(psu, sizeof(bc635pci_statics));

	return (OK);	// according to the dldd template ?
}


// ------------------------------------------------------------------------ 
// Name: bc635pci_open
// 
// Description: The open routine is called upon an application open of the
// device.
// 
// Input: statics pointer, device number, file structure pointer.
// 
// Returns: OK or SYSERR
// ------------------------------------------------------------------------
int	bc635pci_open(	bc635pci_statics	*psu, 
					int					dev,
					struct file			*f)

{

#ifdef DEBUG_DRVR
	cprintf("bc635pci_open() OPEN\n");
#endif

	// during debug it was noted that even though the install failed the kernel
	// would still access the device specific entry points but...
	if (psu == NULL)		// install returns a NULL pointer if the device does not exist
	{
		pseterr (ENXIO);	// device does not exist.
		return(SYSERR);
	}

	// we don't care how they open the device because it's FREAD and FWRITE
	// there is no minor device initialization required

	if(psu->state != BC_PCI_OPEN)
	{
		psu->state = BC_PCI_OPEN;
	}

// set the kernel thread 

	stsetprio (psu->thread_id, (( _getpriority() * 2) + 3));

	return (OK);
}


// ------------------------------------------------------------------------ 
// Name: bc635pci_close
// 
// Description: The close routine is called upon an application close of the
// device.
// 
// Input: statics pointer, file structure pointer
// 
// Returns: OK or SYSERR
// ------------------------------------------------------------------------
int	bc635pci_close(	bc635pci_statics	*psu,
					struct file			*f)

{

#ifdef DEBUG_DRVR
	cprintf("bc635pci_close() CLOSE\n");
#endif
	// during debug it was noted that even though the install failed the kernel
	// would still access the device specific entry points but...
	if (psu == NULL)		// install returns a NULL pointer if the device does not exist
	{
		pseterr (EBADF);	// device does not exist / fd is not a valid file descriptor
		return(SYSERR);
	}


	if (psu->state == BC_PCI_OPEN)
		psu->state = BC_PCI_INSTALL;

//	set the kernel thread priority to a base number
	stsetprio (psu->thread_id, psu->thread_prio);

	return (OK);
}


// ------------------------------------------------------------------------ 
// Name: bc635pci_read
// 
// Description: The read routine is called upon an application read from the
// device.
// 
// Input: statics pointer, file structure pointer, buffer pointer, count of 
// bytes to read.
// 
// Returns: count of the bytes actually read.
// ------------------------------------------------------------------------
int	bc635pci_read(	bc635pci_statics	*psu,
					struct file			*f,
					char				*buff,
					register int		icnt)
{

	int	rcnt = 0;					// length of GPS packet

#ifdef DEBUG_DRVR
	cprintf("bc635pci_read() READ\n");
#endif
	// during debug it was noted that even though the install failed the kernel
	// would still access the device specific entry points but...
	if (psu == NULL)		// install returns a NULL pointer if the device does not exist
	{
		pseterr (EBADF);	// device does not exist / fd is not a valid file descriptor
		return(SYSERR);
	}

	if (psu->pGps[0] == 0)			// check availabiltiy of GPS packet
	{
		pseterr (EIO);
		return(SYSERR);
	}

	rcnt = (int)psu->pGps[0] + 1;	// read length of GPS Packet
									// from PCI register file
	if (rcnt > STFP_MAX_READ)		// check length of GPS packet
	{
		pseterr (EIO);
		return(SYSERR);
	}

	if (rcnt > icnt)				// buffer not big enough
	{
		pseterr (EMSGSIZE);			// Message to long
		return(SYSERR);
	}

	if (wbounds((unsigned long)buff) < rcnt)	// check pointer validity 
	{	// not enough bytes to the next boundary of non-writable memory
		pseterr (EINVAL);
		return(SYSERR);
	}

	bcopy(psu->pGps, buff, rcnt) ;	// copy / move data to user address space
	return (rcnt);					// return number of bytes in packet
}	// end bc635pci_read driver entry point

// ------------------------------------------------------------------------ 
// Name: bc635pci_write
// 
// Description: The write routine is called upon an application write to the
// device.
// 
// Input: statics pointer, file structure pointer, buffer pointer, count of 
// bytes to write.
// 
// Returns: count of the bytes actually written.
// ------------------------------------------------------------------------
int bc635pci_write(	bc635pci_statics	*psu,
					struct file			*f,
					char				*buff,
					register int		count)
{
int				retval = 0;
unsigned long	argument;
// struct timespec time;
int				time;


#ifdef DEBUG_DRVR
	cprintf("bc635pci_write() WRITE\n");
#endif

	// during debug it was noted that even though the install failed the kernel
	// would still access the device specific entry points but...
	if (psu == NULL)		// install returns a NULL pointer if the device does not exist
	{
		pseterr (EBADF);	// device does not exist / fd is not a valid file descriptor
		return(SYSERR);
	}

	/* Check length of the written GPS packet */

	if(count > STFP_MAX_WRITE)
	{
		pseterr(EFBIG);
		return(SYSERR);
	}

	if (rbounds((unsigned long)buff) < count)	// check pointer validity 
	{	// not enough bytes to the next boundary of non-readable memory
		pseterr (EINVAL);
		return(SYSERR);
	}

	bcopy(buff,psu->pIn,count) ;	// copy / move data from user address space

	/* handle a response if there is one */

	if ((psu->pIn[0] == COM_MANREQGPSPACKET) && (psu->pIn[2] != 0))
	{
		/* do not sleep if INT_PACKET signal enabled */

		if(psu->intMaskSig & INT_PACKET)
		{
			retval = sendCommand(psu);
		}
		else
		{
			argument = psu->pRegs->mask & ~INT_PACKET;		// ANDEQUAL
			psu->pRegs->mask =		argument;				// set AND
			psu->pRegs->intstat =	INT_PACKET;
			psu->pRegs->ack =		ACK_PACKET;
			retval = sendCommand(psu);

			/* sleep if packet is not immediately available */
#ifdef DEBUG_DRVR
	cprintf("bc635pci_write() sleep if packet is not immediately available\n");
#endif


			swait(&psu->mutex, SEM_SIGIGNORE);				// enter critical code region

			if(!(psu->pRegs->intstat & INT_PACKET))
			{
				argument = psu->pRegs->mask | INT_PACKET;	
				psu->pRegs->mask = argument;				// set OREQUAL
				ssignal(&psu->mutex);		// leave critical code region		

				time = 5 * (1000 / 10);					// 10ms ticks for 5 seconds
				if(tswait(&psu->cv, SEM_SIGABORT, time) == 0)
				{
					/* if timeout */
					pseterr(EIO);
					return(SYSERR);
				}

				swait(&psu->mutex, SEM_SIGIGNORE);				// enter critical code region
				argument = psu->pRegs->mask & ~INT_PACKET;		// ANDEQUAL
				psu->pRegs->mask = argument;
				ssignal(&psu->mutex);			// leave critical code region		
			}
			else
				ssignal(&psu->mutex);			// leave critical code region		
		}
	}
	else
		retval = sendCommand(psu);

	//	return either the number of bytes written (count) or SYSERR.

	if (retval == SYSERR)
	{
		return (retval);
	}
	else
	{
		return (count);
	}
}	// end bc635pci_write driver entry point


// ------------------------------------------------------------------------ 
// Name: bc635pci_ioctl
// 
// Description: The ioctl is called upon the application ioctl to the device
// 
// Input: statics pointer, file structure pointer, command, argument to command
// 
// Returns: OK or SYSERR
// ------------------------------------------------------------------------
int	bc635pci_ioctl(	bc635pci_statics	*psu,
					struct file			*f,
					int					cmd,
					char				*args)
{
struct	stfp_timeval	Tv;
struct	stfp_time		Tm;
struct	stfp_dec_tm		dec;
struct	stfp_version	Ver;
struct	periodic		Per;
struct	tcgenoffset		tcgoff;
struct	loctmoffset		loctoff;
struct	leapseconds		lpsec;
struct	getdata_t		getdata;
unsigned char			*getdataptr;
unsigned long			argument;
// struct timespec 		time;
int						time;
int						retval = 0;
int						get_data_len;
int						i;


#ifdef DEBUG_DRVR
	cprintf("bc635pci_ioctl() IOCTL\n");
#endif

	// during debug it was noted that even though the install failed the kernel
	// would still access the device specific entry points but...
	if (psu == NULL)		// install returns a NULL pointer if the device does not exist
	{
		pseterr (ENODEV);	// device does not exist / fd does not refer to a valid device
		return(SYSERR);
	}

	retval = OK;	// assume ok

	switch(cmd)
	{
		case TIMEREQUEST:
			psu->pRegs->timereq = 0;
			break;

		case EVENTREQUEST:
			psu->pRegs->eventreq = 0;
			break;

		case RDTIME:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_time))
			{	// not enough bytes to the next boundary of non-writable memory
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				rdTime(psu, psu->pRegs->time, psu->pYear, &Tm);		// read time from board
				// copy the time from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tm, (caddr_t)args, sizeof(struct stfp_time));
			}
			break;

		case RDEVENT:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_time))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				rdTime(psu, psu->pRegs->event, psu->pYear, &Tm);	// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tm, (caddr_t)args, sizeof(struct stfp_time));
			}
			break;

		case RDTIMEREQ:									// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_time))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pRegs->timereq = 0;
				rdTime(psu, psu->pRegs->time, psu->pYear, &Tm);		// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tm, (caddr_t)args, sizeof(struct stfp_time));
			}
			break;

		case RDEVENTREQ:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_time))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pRegs->eventreq = 0;
				rdTime(psu, psu->pRegs->event, psu->pYear, &Tm);	// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tm, (caddr_t)args, sizeof(struct stfp_time));
			}
			break;

		case RDTIMETV:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				rdTimeTv(psu, psu->pRegs->time, &Tv);				// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tv, (caddr_t)args, sizeof(struct stfp_timeval));
			}
			break;

		case RDEVENTTV:									// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				rdTimeTv(psu, psu->pRegs->event, &Tv);				// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tv, (caddr_t)args, sizeof(struct stfp_timeval));
			}
			break;

		case RDTIMETVREQ:								// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pRegs->timereq = 0;
				rdTimeTv(psu, psu->pRegs->time, &Tv);				// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tv, (caddr_t)args, sizeof(struct stfp_timeval));
			}
			break;

		case RDEVENTTVREQ:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pRegs->eventreq = 0;
				rdTimeTv(psu, psu->pRegs->event, &Tv);				// read data from board
				// copy the data from the kernel space to the user space
				// assume args is pointer to time structure
				bcopy((caddr_t)&Tv, (caddr_t)args, sizeof(struct stfp_timeval));
			}
			break;

		case CAPUNLOCK:
			psu->pRegs->unlock = 0;
			break;

		case RDINTSTAT:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = psu->pRegs->intstat & INT_BITMASK;
				bcopy((caddr_t)&argument, (caddr_t)args, sizeof(unsigned long));
			}
			break;

		case CLRINTSTAT:
			psu->pRegs->intstat = (unsigned long)args & ~psu->intMaskSig;
			break;
		
		case RDCONTROL:									// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = psu->pRegs->control;
				bcopy((caddr_t) &argument, (caddr_t)args, sizeof(unsigned long));
			}
			break;	
			
		case SETCONTROL:								// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy (args, &argument, sizeof(unsigned long));
				psu->pRegs->control = argument;
			}
			break;	

		case WRSTROBE:
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(struct stfp_time))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t)&Tm, sizeof(struct stfp_time)); // read user data
				argument = psu->pRegs->control & ~CTRL_STREN;
				psu->pRegs->control = argument;	// write control (AND EQUAL)

				argument = (unsigned long)((Tm.tm.tm_hour << 16) + (Tm.tm.tm_min << 8) + Tm.tm.tm_sec) & 0xffffff;
				psu->pRegs->mijstrb = argument;	// write major strobe time
				argument =  (unsigned long)(Tm.usec & 0xfffff);
				psu->pRegs->minstrb = argument;	// write minor strobe time

				argument = psu->pRegs->control | CTRL_STREN;
				psu->pRegs->control = argument;	// write control (OR EQUAL)
			}
			break;

		case WRSTROBETV:								// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t)&Tv,	sizeof(struct stfp_timeval));
		
				argument = psu->pRegs->control & ~CTRL_STREN;
				psu->pRegs->control = argument;	// write control (AND EQUAL)

				psu->pRegs->mijstrb = (unsigned long)Tv.tv.tv_sec;	// majjor strobe time
																// minor strobe tune
				psu->pRegs->minstrb = (unsigned long)(Tv.tv.tv_usec & 0xfffff);

				argument = psu->pRegs->control | CTRL_STREN;
				psu->pRegs->control = argument;	// write control (OR EQUAL)
			}
			break;

		case SETTIME:									// Tested by bc635pci.c DEMO	
			/* Binary Time Format */
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				//	read date from user space
				bcopy (args, &argument, sizeof(unsigned long));
				psu->pIn[0] = COM_SETTIME;

				// what is the format of the input buffer for setting the time
				// args is a type char * by default (to the user it is void *)

				psu->pIn[1] = (unsigned char)(argument >> 24);
				psu->pIn[2] = (unsigned char)(argument >> 16);
				psu->pIn[3] = (unsigned char)(argument >> 8);
				psu->pIn[4] = (unsigned char)argument;
				retval = sendCommand(psu);
				// now we know what drv_usectohs means 80000 == 800ms
				// delay(drv_usectohz(800000)); /* Wait 800ms */
				usec_sleep((unsigned long)800000);		// Wait 800ms
			}
			break;

		case SETDECTIME:								// Tested by bc635pci.c DEMO
			/* Decimal Time Format */
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(struct stfp_dec_tm))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t)&dec,	sizeof(struct stfp_dec_tm));
				psu->pIn[0] = COM_SETTIME;
				psu->pIn[1] = (unsigned char)(dec.dec_year >> 8);
				psu->pIn[2] = (unsigned char)dec.dec_year;
				psu->pIn[3] = (unsigned char)(dec.dec_day >> 8);
				psu->pIn[4] = (unsigned char)dec.dec_day;
				psu->pIn[5] = (unsigned char)dec.dec_hour;
				psu->pIn[6] = (unsigned char)dec.dec_min;
				psu->pIn[7] = (unsigned char)dec.dec_sec;
				retval = sendCommand(psu);
				usec_sleep((unsigned long)800000);		// Wait 800ms
			}
			break;

		case SETTIMETV: /* Get time from tv structure */
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(struct stfp_timeval))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{	// dereference pointer args - read in data
				bcopy((caddr_t)args, (caddr_t) &Tv,sizeof(struct stfp_timeval));
				psu->pIn[0] = COM_SETTIME;
				psu->pIn[1] = (unsigned char)(Tv.tv.tv_sec >> 24);
				psu->pIn[2] = (unsigned char)(Tv.tv.tv_sec >> 16);
				psu->pIn[3] = (unsigned char)(Tv.tv.tv_sec >> 8);
				psu->pIn[4] = (unsigned char)Tv.tv.tv_sec;
				retval = sendCommand(psu);
				usec_sleep((unsigned long)800000);		// Wait 800ms
			}
			break;


		case SETYEAR:									// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy (args, &argument, sizeof(unsigned long));	// dereferenc pointer

				psu->pIn[0] = COM_SETYEAR;
				psu->pIn[1] = (unsigned char)(argument >> 8);
				psu->pIn[2] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case SETINTSIGNAL:
			// check validity of pointer args before using
			// argument is the signal the process is going to wait on.
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
#ifdef DEBUG_DRVR
		cprintf ("IOCTL() SETINTSIGNAL rbounds (%d) = %d\n",
				(unsigned long)args, 
				rbounds((unsigned long)args)); 
#endif
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				// dereference pointer
				bcopy (args, &argument, sizeof(unsigned long));
				swait(&psu->mutex, SEM_SIGIGNORE);			// mutex enter

				psu->intMaskSig = argument & INT_BITMASK;

//	Questiuons what does "proc_ref()" do ? similiar to getpid()

				if (psu->intMaskSig)
					psu->proc_id = getpid();
				else 
					psu->proc_id = 0;
//  no equ in LynxOS - not needed in LynxOs		proc_unref(pref);

				psu->pRegs->intstat = psu->intMaskSig;

				if(psu->intMaskSig & INT_PACKET)
					psu->pRegs->ack = ACK_PACKET;

				psu->pRegs->mask = (psu->pRegs->mask & INT_PACKET) | psu->intMaskSig;

				ssignal(&psu->mutex);			// mutex exit
			}
			break;

		case RDINTSIGNAL:
			// check validity of pointer args before using
			if (wbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				swait(&psu->mutex, SEM_SIGIGNORE);			// mutex enter
				argument = psu->intServiced & INT_BITMASK;
				psu->intServiced = 0;
				ssignal(&psu->mutex);						// mutex exit
				bcopy((caddr_t) &argument, (caddr_t)args,	sizeof(unsigned long));
			}
			break;

		case REQGPSPACKET:
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_REQGPSPACKET;
				psu->pIn[1] = args[0];

				/* do not sleep if INT_PACKET signal enabled */

				if(psu->intMaskSig & INT_PACKET)
				retval = sendCommand(psu);
				else
				{
					argument = psu->pRegs->mask & ~INT_PACKET;		// STFP_REG_ANDEQUAL	
					psu->pRegs->mask = argument;

					psu->pRegs->intstat = INT_PACKET;
					psu->pRegs->ack = ACK_PACKET;
					retval = sendCommand(psu);

					// sleep if packet is not immediately available

					swait(&psu->mutex, SEM_SIGIGNORE);				// mutex_enter

					if(!(psu->pRegs->intstat & INT_PACKET))
					{
						argument = psu->pRegs->mask | INT_PACKET;	// STFP_REG_OREQUAL
						psu->pRegs->mask = argument;
						ssignal(&psu->mutex);						// mutex_exit

						time = 5 * (1000 / 10);					// 10ms ticks for 5 seconds
						if(tswait(&psu->cv, SEM_SIGABORT, time) == 0)
						{
						/* if timeout */
							retval = SYSERR;
							pseterr(EFAULT);
							break;
						}
						swait(&psu->mutex, SEM_SIGIGNORE);			// mutex_enter
						argument = psu->pRegs->mask & ~INT_PACKET;	// STFP_REG_ANDEQUAL
						psu->pRegs->mask = argument;
						ssignal(&psu->mutex);		// mutex_exit
					}
					else
						ssignal(&psu->mutex);		// mutex_exit
				}	// end else
			}	// bounds test
			break;

		case CONTROLTIMEBASE:							// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(int))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy (args, &i, sizeof (int));	// dereference user data
				switch (i)
				{
					case CLOCK_INTERNAL:
						psu->pIn[0] = COM_SELCLOCK;
						psu->pIn[1] = SELCLOCK_INT;
						retval = sendCommand(psu);
						break;
					case CLOCK_EXTERNAL:
						psu->pIn[0] = COM_SELCLOCK;
						psu->pIn[1] = SELCLOCK_EXT;
						retval = sendCommand(psu);
						break;
					case JAMSYNC_ENABLE:
						psu->pIn[0] = COM_JAMSYNCENABLE;
						psu->pIn[1] = JAMSYNC_ENABLE;
						retval = sendCommand(psu);
						break;
					case JAMSYNC_DISABLE:
						psu->pIn[0] = COM_JAMSYNCENABLE;
						psu->pIn[1] = JAMSYNC_DISABLE;
						retval = sendCommand(psu);
						break;
					case DISCIPLINE_ENABLE:
						psu->pIn[0] = COM_DISCENABLE;
						psu->pIn[1] = DISC_EN;
						retval = sendCommand(psu);
						break;
					case DISCIPLINE_DISABLE:
						psu->pIn[0] = COM_DISCENABLE;
						psu->pIn[1] = DISC_DIS;
						retval = sendCommand(psu);
						break;
					case FORCE_JAMSYNC:
						psu->pIn[0] = COM_FORCEJAMSYNC;
						retval = sendCommand(psu);
						break;
					case CLEAR_USEC:
						break;
					default:
							retval = SYSERR;
							pseterr(EINVAL);
						break;
				}	// end switch
			}		// end bounds test
			break;

		case SETDAC:									// Tested by bc635pci.c DEMO
			// check validity of pointer args before using
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				// dereference user data
				bcopy (args, &argument, sizeof(unsigned long));
				psu->pIn[0] = COM_SETDAC;
				psu->pIn[1] = (unsigned char)(argument >> 8);
				psu->pIn[2] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case RDDAC:
			if (wbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_GETDATA;
				psu->pIn[1] = GETDATA_DAC;
				retval = sendCommand(psu);
				if( retval == 0)	// if sendCommand successful continue - else break
				{
					argument =	(unsigned long)(psu->pOut[1] << 8) & 0xff00;
					argument |= (unsigned long)(psu->pOut[2] & 0x00ff);
					bcopy((caddr_t)&argument, (caddr_t)args, sizeof(unsigned long));
				}
			}
			break;

		case SETTIMINGOFFSET:							// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
			// dereference user data
				bcopy(args, &argument, sizeof(unsigned long));
				psu->pIn[0] = COM_SETTIMINGOFFSET;
				psu->pIn[1] = (unsigned char)(argument >> 24);
				psu->pIn[2] = (unsigned char)(argument >> 16);
				psu->pIn[3] = (unsigned char)(argument >> 8);
				psu->pIn[4] = (unsigned char)(argument);
				retval = sendCommand(psu);
			}
			break;
	
		case SELTIMINGMODE:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_SETTIMINGMODE;
				psu->pIn[1] = args[0];
				retval = sendCommand(psu);
				usec_sleep((unsigned long)1000000);		// wait 1sec to settle
			}
			break;

		case SELTIMEFORMAT:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_SETTIMEFORMAT;
				psu->pIn[1] = args[0];
				retval = sendCommand(psu);
				usec_sleep((unsigned long)1000000);		// wait 1sec to settle
			}
			break;

		case SELTCFORMAT:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_SETTCFORMAT;
				psu->pIn[1] = args[0];
				retval = sendCommand(psu);
				usec_sleep((unsigned long)1000000);		// wait 1sec to settle
			}
			break;

		case SELTCMOD:									// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_SETTCMOD;
				psu->pIn[1] = args[0];
				retval = sendCommand(psu);
				usec_sleep((unsigned long)1000000);		// wait 1sec to settle
			}
			break;

		case SETPERIODIC:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(struct periodic))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t) &Per, sizeof(struct periodic));

				psu->pIn[0] = COM_SETPERIODIC;
				psu->pIn[1] = (unsigned char)Per.sync1pps;
				psu->pIn[2] = (unsigned char)(Per.n1 >> 8);
				psu->pIn[3] = (unsigned char)(Per.n1);
				psu->pIn[4] = (unsigned char)(Per.n2 >> 8);
				psu->pIn[5] = (unsigned char)(Per.n2);

				retval = sendCommand(psu);
			}
			break;

		case CONTROLEVENT:
			if (rbounds((unsigned long)args) < sizeof(int))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				// dereference args
				bcopy(args, &i, sizeof(int));
				switch (i)
				{
					case EVENT_RISING:
						// STFP_REG_ANDEQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_EVSENSE;
						break;
					case EVENT_FALLING:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_EVSENSE;
						break;
					case CAPLOCK_ENABLE:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_LOCKEN;
						break;
					case CAPLOCK_DISABLE:
						// STFP_REG_ANDEQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_LOCKEN;
						break;
					case EVCAP_ENABLE:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_EVENTEN;
						break;
					case EVCAP_DISABLE:
						// STFP_REG_ANDEQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_EVENTEN;
						break;
					case EVCAP_PERIODIC:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_EVSOURCE;
						break;
					case EVCAP_EVENT:
						// STFP_REG_ANDEQUAL(
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_EVSOURCE;
						break;
					default:
						pseterr (EINVAL);
						retval = SYSERR;
				}	// end switch 
			}		// end bounds test
			break;

		case CONTROLSTROBE:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(int))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				// dereference args
				bcopy(args, &i, sizeof(int));
				switch (i)
				{
					case STROBE_SECUS:
						// STFP_REG_ANDEQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_STRMODE;
						break;
					case STROBE_USONLY:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_STRMODE;
						break;
					case STROBE_ENABLE:
						// STFP_REG_OREQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument | CTRL_STREN;
						break;
					case STROBE_DISABLE:
						// STFP_REG_ANDEQUAL
						argument = psu->pRegs->control;
						psu->pRegs->control = argument & ~CTRL_STREN;
						break;
					default:
						pseterr (EINVAL);
						retval = SYSERR;
						break;
				}	// end switch
			}		// end bounds test
			break;

		case SELFREQUENCYOUT:							// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				// dereference args
				i = (int)*(args);
				switch (i)
				{
					case FREQ_10MHZ:
						argument = psu->pRegs->control;
						psu->pRegs->control = ((argument & ~CTRL_FMASK) | CTRL_F10);
						break;
					case FREQ_5MHZ:
						argument = psu->pRegs->control;
						psu->pRegs->control = ((argument & ~CTRL_FMASK) | CTRL_F5);
						break;
					case FREQ_1MHZ:
						argument = psu->pRegs->control;
						psu->pRegs->control = ((argument & ~CTRL_FMASK) | CTRL_F1);
						break;
					default:
						pseterr (EINVAL);
						retval = SYSERR;
						break;
				}	// end switch
			}		// end bounds test
			break;

		case SOFTWARERESET:								// Tested by bc635pci.c DEMO
			psu->pIn[0] = COM_SOFTRESET;
			psu->pIn[1] = SW_RESET;
			retval = sendCommand(psu);
			usec_sleep((unsigned long)1000000);			// wait 1sec to settle
			break;

		case SETTCOUTFMT:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_SETTCOUTFMT;
				psu->pIn[1] = (unsigned char)*(args);
				retval = sendCommand(psu);
			}
			break;


		case SETGENTMOFFSET:							// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(struct tcgenoffset))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{

				bcopy((caddr_t)args, (caddr_t)&tcgoff, sizeof(struct tcgenoffset));
				psu->pIn[0] = COM_SETGENTMOFF;
				psu->pIn[1] = (tcgoff.genoff >> 8);
				psu->pIn[2] = (tcgoff.genoff);
				psu->pIn[3] = (tcgoff.genflg);
				retval = sendCommand(psu); 
			}
			break;

		case SETLOCTMOFFSET:							// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(struct loctmoffset))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{

				bcopy((caddr_t)args, (caddr_t) &loctoff, sizeof(struct loctmoffset));
				psu->pIn[0] = COM_SETLOCTMOFF;
				psu->pIn[1] =  (loctoff.locoff >> 8);
				psu->pIn[2] =  loctoff.locoff;
				psu->pIn[3] =  loctoff.locflg;
				retval = sendCommand(psu); 
			}
			break;
			
		case SETLEAPSECEVENT:							// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(struct leapseconds))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t)&lpsec, sizeof(struct leapseconds));
				psu->pIn[0] = COM_SETLEAPSEC;
				psu->pIn[1] = lpsec.ls_flag;
				psu->pIn[2] = (unsigned char)(lpsec.leap_tm >> 24);
				psu->pIn[3] = (unsigned char)(lpsec.leap_tm >> 16);
				psu->pIn[4] = (unsigned char)(lpsec.leap_tm >> 8);
				psu->pIn[5] = (unsigned char)lpsec.leap_tm;
				retval = sendCommand(psu);
			}
			break;

		case SETDISGAIN:
			if (rbounds((unsigned long)args) < sizeof(unsigned short))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = 0;
				bcopy (args, &argument, sizeof(unsigned short));
				psu->pIn[0] = COM_SETDISCGAIN;
				psu->pIn[1] = (unsigned char)(argument >> 8);
				psu->pIn[2] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case SYNCRTC:									// Tested by bc635pci.c DEMO
			psu->pIn[0] = COM_SYNCRTC;
			retval = sendCommand(psu);
			break;

		case DISCBATT:									// Tested by bc635pci.c DEMO
			psu->pIn[0] = COM_DISCBATTRTC;
			retval = sendCommand(psu);
			break;

		case SETCLKVAL:									// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned long))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy (args, &argument, sizeof(unsigned long));
				psu->pIn[0] = COM_SETCLKVAL;
				psu->pIn[1] = (unsigned char)(argument >> 24);
				psu->pIn[2] = (unsigned char)(argument >> 16);
				psu->pIn[3] = (unsigned char)(argument >> 8);
				psu->pIn[4] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;
		
		case SETGPSTMFMT:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = 0;
				bcopy (args, &argument, sizeof(unsigned char));
				psu->pIn[0] = COM_GPSTIMEFORMAT;
				psu->pIn[1] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case SETGPSMDFLG:
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = 0;
				bcopy (args, &argument, sizeof(unsigned char));
				psu->pIn[0] = COM_SETGPSMDFLG;
				psu->pIn[1] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case SETLOCTMFLG:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = 0;
				bcopy (args, &argument, sizeof(unsigned char));
				psu->pIn[0] = COM_SETLOCTMFLG;
				psu->pIn[1] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;

		case SETYRINCFLG:								// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(unsigned char))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				argument = 0;
				bcopy (args, &argument, sizeof(unsigned char));
				psu->pIn[0] =COM_SETYRINCFLG;
				psu->pIn[1] = (unsigned char)argument;
				retval = sendCommand(psu);
			}
			break;


		case RDFIRMVERSION:
			// check pointer validity 
			if (wbounds((unsigned long)args) < sizeof(struct stfp_version))
			{	// not enough bytes to the next boundary of non-writable memory
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				psu->pIn[0] = COM_GETDATA;
				psu->pIn[1] = GETDATA_VERSION;
				retval = sendCommand(psu);
				if(retval == 0)	// if sendCommand successful - else break
				{
					Ver.vmajor = psu->pOut[1];
					Ver.vminor = psu->pOut[2];
					Ver.month = psu->pOut[3];
					Ver.day = psu->pOut[4];
					Ver.year = (unsigned short)((psu->pOut[5] << 8) | psu->pOut[6]);
					bcopy((caddr_t) &Ver, (caddr_t)args,	sizeof(struct stfp_version));
				}	// end if command successful
			}
			break;

		case GETDATA:									// Tested by bc635pci.c DEMO
			if (rbounds((unsigned long)args) < sizeof(struct getdata_t))
			{
				pseterr (EINVAL);
				retval = SYSERR;
			}
			else
			{
				bcopy((caddr_t)args, (caddr_t) &getdata,	sizeof(struct getdata_t));

				psu->pIn[0] = COM_GETDATA;
				psu->pIn[1] = getdata.arg;
				retval = sendCommand(psu);

				if(retval == 0)	// if sendCommand successful - else break
				{
					/* get the number of bytes to be transferred */
					switch(getdata.arg)
					{
						case GETDATA_MODE:
							get_data_len=sizeof(getdata.data.tmode);
							break;

						case GETDATA_TFORMAT:
							get_data_len=sizeof(getdata.data.tformat);
							break;

						case GETDATA_YEAR:
							get_data_len=sizeof(getdata.data.year);
							break;

						case GETDATA_PER:
							get_data_len=sizeof(getdata.data.period);
							break;

						case GETDATA_TCFMT:
							get_data_len=sizeof(getdata.data.tcformat);
							break;

						case GETDATA_TCMOD:
							get_data_len=sizeof(getdata.data.tcmod);
							break;

						case GETDATA_TOFF:
							get_data_len=sizeof(getdata.data.offset);
							break;

						case GETDATA_UTCINFO:
							get_data_len=sizeof(getdata.data.utc);
							break;
					
						case GETDATA_TCOUTFMT:
							get_data_len=sizeof(getdata.data.tcoutfmt);
							break;

						case GETDATA_TCGENOFF:
							get_data_len=sizeof(getdata.data.genoff);
							break;

						case GETDATA_LOCTMOFF:
							get_data_len=sizeof(getdata.data.locoff);
							break;

						case GETDATA_LEAPSEC:
							get_data_len=sizeof(getdata.data.leapsec);
							break;

						case GETDATA_FWVER:
							get_data_len=sizeof(getdata.data.version);
							break;

						case GETDATA_CLKSRC:
							get_data_len=sizeof(getdata.data.clksrc);
							break;

						case GETDATA_JAMSC:
							get_data_len=sizeof(getdata.data.jsctl);
							break;

						case GETDATA_OSCCTL:
							get_data_len=sizeof(getdata.data.ocsctl);
							break;

						case GETDATA_DAVAL:
							get_data_len=sizeof(getdata.data.davalue);
							break;
						
						case GETDATA_BATTSTAT:
							get_data_len=sizeof(getdata.data.battstat);
							break;

						case GETDATA_CLKVAL:
							get_data_len=sizeof(getdata.data.clkval);
							break;

						case GETDATA_DTFW:
							get_data_len=sizeof(getdata.data.dtfw);
							break;

						case GETDATA_ASSEMB:
							get_data_len=sizeof(getdata.data.assemb);
							break;

						case GETDATA_HWFAB:
							get_data_len=sizeof(getdata.data.hwfab);
							break;

						case GETDATA_TFPMODEL:
							get_data_len=sizeof(getdata.data.tfpm);
							break;

						case GETDATA_SERIAL:
							get_data_len=sizeof(getdata.data.serialnum);
							break;

						default:
							get_data_len=0;
							pseterr (EFAULT);
							retval = SYSERR;
					}	// end switch number of bytes to be transferred

					if (retval == OK)
					{
						/* Skip over response byte. Only return data */
						getdataptr=(u_char *)&(getdata.data.tmode);

						for(i=0; i<get_data_len; i++)
						{
							getdataptr[i] = psu->pOut[i+1];
						}
						bcopy((caddr_t) &getdata, (caddr_t)args, sizeof(struct getdata_t));
					}
				} // end if send command successful
			}
			break;

		default:
			pseterr (ENOTTY);
			retval = SYSERR;
			break;
	}	// end switch (cmd)
	
	return (retval);

}	// end bc635pci_ioctl() entry point code


// ------------------------------------------------------------------------ 
// Name: bc635pci_select
// 
// Description: The select is called upon the application invokation of the
// select system call.
// 
// Input: statics pointer, file structure pointer, which command, select 
// structure pointer.
// 
// Returns: OK or SYSERR
// ------------------------------------------------------------------------

int	bc635pci_select	(	bc635pci_statics		*psu,
						struct file				*f, 
						int						which,
						struct sel				*ffs	)

{
	// no needfor the select function for the board - there is always data to read.
	switch (which)
	{
		case SREAD:
			pseterr(EINVAL);
			break;

		case SWRITE:
			pseterr(EINVAL);
			break;

		case SEXCEPT:
			pseterr(EINVAL);
			break;

	default:
		pseterr(EINVAL);
		break;
	}
	return (SYSERR);
}

// TJL 30-APR-02	to make a static driver this structure goses away and
//					entry points are in the bc635pci.cfg file.

static struct dldd entry_points = {
						bc635pci_open,
						bc635pci_close,
						bc635pci_read,
						bc635pci_write,
						bc635pci_select,
						bc635pci_ioctl,
						bc635pci_install,
						bc635pci_uninstall, 	
						(char *)0 
				   };


// -------------------------------  EOF ----------------------------------
